@@ -41,6 +41,7 @@
 #include <mntent.h>
 #include <fcntl.h>
 #include <limits.h>
+#include <stdbool.h>
 #include <paths.h>
 #include <libgen.h>
 #include <time.h>
@@ -124,6 +125,7 @@
 #define CRED_USER        1
 #define CRED_PASS        2
 #define CRED_DOM         4
+#define CRED_PASS2		 5
 
 /*
  * Values for parsing command line options.
@@ -163,6 +165,7 @@
 #define OPT_BKUPGID    31
 #define OPT_NOFAIL     32
 #define OPT_SNAPSHOT   33
+#define OPT_PASS2	   34
 
 #define MNT_TMP_FILE "/.mtab.cifs.XXXXXX"
 
@@ -185,15 +188,18 @@ struct parsed_mount_info {
 	char domain[MAX_DOMAIN_SIZE + 1];
 	char username[MAX_USERNAME_SIZE + 1];
 	char password[MOUNT_PASSWD_SIZE + 1];
+	char password2[MOUNT_PASSWD_SIZE + 1];
 	char addrlist[MAX_ADDR_LIST_LEN];
 	unsigned int got_user:1;
 	unsigned int got_password:1;
+	unsigned int got_password2:1;
 	unsigned int fakemnt:1;
 	unsigned int nomtab:1;
 	unsigned int verboseflag:1;
 	unsigned int nofail:1;
 	unsigned int got_domain:1;
 	unsigned int is_krb5:1;
+	unsigned int is_noauth:1;
 	uid_t sudo_uid;
 };
 
@@ -294,7 +300,7 @@ static int mount_usage(FILE * stream)
 	fprintf(stream,
 		"\n\tsign,seal,fsc,snapshot=<token|time>,nosharesock,persistenthandles,");
 	fprintf(stream,
-		"\n\tresilienthandles,rdma,vers=<smb_dialect>,cruid");
+		"\n\tresilienthandles,rdma,vers=<smb_dialect>,cruid,password2=<alt password>");
 	fprintf(stream,
 		"\n\nOptions not needed for servers supporting CIFS Unix extensions");
 	fprintf(stream,
@@ -330,21 +336,31 @@ static int mount_usage(FILE * stream)
  * end up getting confused for option delimiters. Copy password into pw
  * field, turning any commas into double commas.
  */
-static int set_password(struct parsed_mount_info *parsed_info, const char *src)
+static int
+set_password(struct parsed_mount_info *parsed_info, const char *src,
+			   const int which_pass)
 {
-	char *dst = parsed_info->password;
+	int is_pass2 = which_pass == OPT_PASS2 || which_pass == CRED_PASS2;
+
+	char *dst = is_pass2 ?
+				parsed_info->password2 : parsed_info->password;
+	unsigned int pass_length = is_pass2 ?
+					  sizeof(parsed_info->password2) : sizeof(parsed_info->password);
 	unsigned int i = 0, j = 0;
 
 	while (src[i]) {
 		if (src[i] == ',')
 			dst[j++] = ',';
 		dst[j++] = src[i++];
-		if (j > sizeof(parsed_info->password)) {
+		if (j > pass_length) {
 			fprintf(stderr, "Converted password too long!\n");
 			return EX_USAGE;
 		}
 	}
 	dst[j] = '\0';
+	if (is_pass2)
+		parsed_info->got_password2 = 1;
+	else
 	parsed_info->got_password = 1;
 	return 0;
 }
@@ -559,6 +575,9 @@ static int parse_cred_line(char *line, char **target)
 	/* tell the caller which value target points to */
 	if (strncasecmp("user", line, 4) == 0)
 		return CRED_USER;
+	if (strncasecmp("pass2", line, 5) == 0 ||
+		strncasecmp("password2", line, 9) == 0)
+		return CRED_PASS2;
 	if (strncasecmp("pass", line, 4) == 0)
 		return CRED_PASS;
 	if (strncasecmp("dom", line, 3) == 0)
@@ -623,7 +642,12 @@ static int open_cred_file(char *file_name,
 			parsed_info->got_user = 1;
 			break;
 		case CRED_PASS:
-			i = set_password(parsed_info, temp_val);
+			i = set_password(parsed_info, temp_val, CRED_PASS);
+			if (i)
+				goto return_i;
+			break;
+		case CRED_PASS2:
+			i = set_password(parsed_info, temp_val, CRED_PASS2);
 			if (i)
 				goto return_i;
 			break;
@@ -652,10 +676,14 @@ return_i:
 
 static int
 get_password_from_file(int file_descript, char *filename,
-		       struct parsed_mount_info *parsed_info, const char *program)
+		       struct parsed_mount_info *parsed_info, const char *program,
+			   const int which_pass)
 {
 	int rc = 0;
-	char buf[sizeof(parsed_info->password) + 1];
+	int is_pass2 = which_pass == OPT_PASS2;
+	unsigned int pass_length = is_pass2 ?
+					  sizeof(parsed_info->password2) : sizeof(parsed_info->password);
+	char buf[pass_length + 1];
 
 	if (filename != NULL) {
 		rc = toggle_dac_capability(0, 1);
@@ -697,7 +725,7 @@ get_password_from_file(int file_descript, char *filename,
 		goto get_pw_exit;
 	}
 
-	rc = set_password(parsed_info, buf);
+	rc = set_password(parsed_info, buf, which_pass);
 
 get_pw_exit:
 	if (filename != NULL)
@@ -727,6 +755,9 @@ static int parse_opt_token(const char *token)
 	if (strcmp(token, "pass") == 0 ||
 		strcmp(token, "password") == 0)
 		return OPT_PASS;
+	if (strcmp(token, "pass2") == 0 ||
+		strcmp(token, "password2") == 0)
+		return OPT_PASS2;
 	if (strcmp(token, "sec") == 0)
 		return OPT_SEC;
 	if (strcmp(token, "ip") == 0 ||
@@ -830,7 +861,7 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 	 * wide +1 for NULL, and +1 for good measure
 	 */
 	char txtbuf[22];
-	unsigned long long snapshot;
+	unsigned long long snapshot = 0;
 	struct tm tm;
 
 	/* make sure we're starting from beginning */
@@ -902,18 +933,37 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 				parsed_info->got_password = 1;
 				goto nocopy;
 			}
-			rc = set_password(parsed_info, value);
+			rc = set_password(parsed_info, value, OPT_PASS);
+			if (rc)
+				return rc;
+			goto nocopy;
+
+		case OPT_PASS2:
+		if (parsed_info->got_password2) {
+				fprintf(stderr,
+					"password2 specified twice, ignoring second\n");
+				goto nocopy;
+			}
+			if (!value || !*value) {
+				parsed_info->got_password2 = 1;
+				goto nocopy;
+			}
+			rc = set_password(parsed_info, value, OPT_PASS2);
 			if (rc)
 				return rc;
 			goto nocopy;
 
 		case OPT_SEC:
 			if (value) {
-				if (!strncmp(value, "none", 4))
+				if (!strncmp(value, "none", 4)) {
+					parsed_info->is_noauth = 1;
 					parsed_info->got_password = 1;
+					parsed_info->got_password2 = 1;
+				}
 				if (!strncmp(value, "krb5", 4)) {
 					parsed_info->is_krb5 = 1;
 					parsed_info->got_password = 1;
+					parsed_info->got_password2 = 1;
 				}
 			}
 			break;
@@ -1110,6 +1160,7 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 		case OPT_GUEST:
 			parsed_info->got_user = 1;
 			parsed_info->got_password = 1;
+			parsed_info->got_password2 = 1;
 			goto nocopy;
 		case OPT_RO:
 			*filesys_flags |= MS_RDONLY;
@@ -1218,7 +1269,11 @@ nocopy:
 			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 5, "uid=%s", txtbuf);
+		rc = snprintf(out + out_len, word_len + 5, "uid=%s", txtbuf);
+		if (rc < 0) {
+			fprintf(stderr, "Error in creating mount string\n");
+			return EX_SYSERR;
+		}
 		out_len = strlen(out);
 	}
 	if (parsed_info->is_krb5 && parsed_info->sudo_uid) {
@@ -1238,7 +1293,11 @@ nocopy:
 			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 7, "cruid=%s", txtbuf);
+		rc = snprintf(out + out_len, word_len + 7, "cruid=%s", txtbuf);
+		if (rc < 0) {
+			fprintf(stderr, "Error in creating mount string\n");
+			return EX_SYSERR;
+		}
 		out_len = strlen(out);
 	}
 	if (got_gid) {
@@ -1254,7 +1313,11 @@ nocopy:
 			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 5, "gid=%s", txtbuf);
+		rc = snprintf(out + out_len, word_len + 5, "gid=%s", txtbuf);
+		if (rc < 0) {
+			fprintf(stderr, "Error in creating mount string\n");
+			return EX_SYSERR;
+		}
 		out_len = strlen(out);
 	}
 	if (got_bkupuid) {
@@ -1270,7 +1333,11 @@ nocopy:
 			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 11, "backupuid=%s", txtbuf);
+		rc = snprintf(out + out_len, word_len + 11, "backupuid=%s", txtbuf);
+		if (rc < 0) {
+			fprintf(stderr, "Error in creating mount string\n");
+			return EX_SYSERR;
+		}
 		out_len = strlen(out);
 	}
 	if (got_bkupgid) {
@@ -1286,7 +1353,11 @@ nocopy:
 			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 11, "backupgid=%s", txtbuf);
+		rc = snprintf(out + out_len, word_len + 11, "backupgid=%s", txtbuf);
+		if (rc < 0) {
+			fprintf(stderr, "Error in creating mount string\n");
+			return EX_SYSERR;
+		}
 		out_len = strlen(out);
 	}
 	if (got_snapshot) {
@@ -1302,7 +1373,11 @@ nocopy:
 			strlcat(out, ",", MAX_OPTIONS_LEN);
 			out_len++;
 		}
-		snprintf(out + out_len, word_len + 10, "snapshot=%s", txtbuf);
+		rc = snprintf(out + out_len, word_len + 10, "snapshot=%s", txtbuf);
+		if (rc < 0) {
+			fprintf(stderr, "Error in creating mount string\n");
+			return EX_SYSERR;
+		}
 		out_len = strlen(out);
 	}
 
@@ -1381,13 +1456,25 @@ static int get_pw_from_env(struct parsed_mount_info *parsed_info, const char *pr
 	int rc = 0;
 
 	if (getenv("PASSWD"))
-		rc = set_password(parsed_info, getenv("PASSWD"));
+		rc = set_password(parsed_info, getenv("PASSWD"), OPT_PASS);
 	else if (getenv("PASSWD_FD"))
 		rc = get_password_from_file(atoi(getenv("PASSWD_FD")), NULL,
-					    parsed_info, program);
+					    parsed_info, program, OPT_PASS);
 	else if (getenv("PASSWD_FILE"))
 		rc = get_password_from_file(0, getenv("PASSWD_FILE"),
-					    parsed_info, program);
+					    parsed_info, program, OPT_PASS);
+
+	if (rc < 0)
+		return rc;
+
+	if (getenv("PASSWD2"))
+		rc = set_password(parsed_info, getenv("PASSWD2"), OPT_PASS2);
+	else if (getenv("PASSWD2_FD"))
+		rc = get_password_from_file(atoi(getenv("PASSWD2_FD")), NULL,
+					    parsed_info, program, OPT_PASS2);
+	else if (getenv("PASSWD2_FILE"))
+		rc = get_password_from_file(0, getenv("PASSWD2_FILE"),
+					    parsed_info, program, OPT_PASS2);
 
 	return rc;
 }
@@ -1413,6 +1500,8 @@ static struct option longopts[] = {
 	{"domain", 1, NULL, 'd'},
 	{"password", 1, NULL, 'p'},
 	{"pass", 1, NULL, 'p'},
+	{"password2", 1, NULL, 0},
+	{"pass2", 1, NULL, 0},
 	{"credentials", 1, NULL, 'c'},
 	{"port", 1, NULL, 'P'},
 	{"sloppy", 0, NULL, 's'},
@@ -1600,7 +1689,7 @@ add_mtab_exit:
 static int
 del_mtab(char *mountpoint)
 {
-	int len, tmprc, rc = 0;
+	int len, tmprc, rc = 0, tmpfd;
 	FILE *mnttmp, *mntmtab;
 	struct mntent *mountent;
 	char *mtabfile, *mtabdir, *mtabtmpfile = NULL;
@@ -1635,8 +1724,9 @@ del_mtab(char *mountpoint)
 		goto del_mtab_exit;
 	}
 
-	mtabtmpfile = mktemp(mtabtmpfile);
-	if (!mtabtmpfile) {
+	// Use mkstemp instead of mktemp
+	tmpfd = mkstemp(mtabtmpfile);
+	if (tmpfd == -1) {
 		fprintf(stderr, "del_mtab: cannot setup tmp file destination\n");
 		rc = EX_FILEIO;
 		goto del_mtab_exit;
@@ -1646,13 +1736,15 @@ del_mtab(char *mountpoint)
 	if (!mntmtab) {
 		fprintf(stderr, "del_mtab: could not update mount table\n");
 		rc = EX_FILEIO;
+		close(tmpfd);
 		goto del_mtab_exit;
 	}
 
-	mnttmp = setmntent(mtabtmpfile, "w");
+	mnttmp = fdopen(tmpfd, "w");
 	if (!mnttmp) {
 		fprintf(stderr, "del_mtab: could not update mount table\n");
 		endmntent(mntmtab);
+		close(tmpfd);
 		rc = EX_FILEIO;
 		goto del_mtab_exit;
 	}
@@ -1889,8 +1981,11 @@ assemble_mountinfo(struct parsed_mount_info *parsed_info,
 	if (rc)
 		goto assemble_exit;
 
-	if (parsed_info->addrlist[0] == '\0')
+	if (parsed_info->addrlist[0] == '\0') {
 		rc = resolve_host(parsed_info->host, parsed_info->addrlist);
+		if (rc == 0 && parsed_info->verboseflag)
+			fprintf(stderr, "Host \"%s\" resolved to the following IP addresses: %s\n", parsed_info->host, parsed_info->addrlist);
+	}
 
 	switch (rc) {
 	case EX_USAGE:
@@ -1920,7 +2015,8 @@ assemble_mountinfo(struct parsed_mount_info *parsed_info,
 		parsed_info->got_user = 1;
 	}
 
-	if (!parsed_info->got_password) {
+	// no need to prompt for password2
+	if (!parsed_info->got_password && !(parsed_info->flags & MS_REMOUNT)) {
 		char tmp_pass[MOUNT_PASSWD_SIZE + 1];
 		char *prompt = NULL;
 
@@ -1928,7 +2024,7 @@ assemble_mountinfo(struct parsed_mount_info *parsed_info,
 			prompt = NULL;
 
 		if (get_password(prompt ? prompt : "Password: ", tmp_pass, MOUNT_PASSWD_SIZE + 1)) {
-			rc = set_password(parsed_info, tmp_pass);
+			rc = set_password(parsed_info, tmp_pass, OPT_PASS);
 		} else {
 			fprintf(stderr, "Error reading password, exiting\n");
 			rc = EX_SYSERR;
@@ -2238,7 +2334,7 @@ mount_retry:
 		fprintf(stderr, "%s kernel mount options: %s",
 			thisprogram, options);
 
-	if (parsed_info->got_password) {
+	if (parsed_info->got_password && !(parsed_info->is_krb5 || parsed_info->is_noauth)) {
 		/*
 		 * Commas have to be doubled, or else they will
 		 * look like the parameter separator
@@ -2249,6 +2345,13 @@ mount_retry:
 			fprintf(stderr, ",pass=********");
 	}
 
+	if (parsed_info->got_password2 && !(parsed_info->is_krb5 || parsed_info->is_noauth)) {
+		strlcat(options, ",password2=", options_size);
+		strlcat(options, parsed_info->password2, options_size);
+		if (parsed_info->verboseflag)
+			fprintf(stderr, ",password2=********");
+	}
+
 	if (parsed_info->verboseflag)
 		fprintf(stderr, "\n");
 
@@ -2340,6 +2443,7 @@ do_mtab:
 mount_exit:
 	if (parsed_info) {
 		memset(parsed_info->password, 0, sizeof(parsed_info->password));
+		memset(parsed_info->password2, 0, sizeof(parsed_info->password2));
 		munmap(parsed_info, sizeof(*parsed_info));
 	}
 
